import unittest
from tests.extensions.common import *

import pandas as pd


FOLDER = 'test_fluq_extensions/files'

class DTO:
    def __init__(self,num):
        self.num = num
        self.s = 'test'*num
        self.array = Query.en(range(num)).to_list()


class ExtendedMethodsTests(unittest.TestCase):

    def test_text_file(self):
        data = Query.file.text(path(FOLDER,'../sample.txt')).to_list()
        self.assertListEqual(["1","2","","3","4"],data)

    def test_text_io(self):
        os.makedirs(path(FOLDER), exist_ok=True)
        array = ['23','342','1241','','452']
        Query.en(array).feed(fluq.to_text_file(path(FOLDER, 'sample1.txt'), autoflush=True))
        self.assertListEqual(array,Query.file.text(path(FOLDER,'sample1.txt')).to_list())

    def test_zip_io(self):
        os.makedirs(path(FOLDER), exist_ok=True)
        array = ['23', '342', '1241', '', '452']
        Query.en(array).feed(fluq.to_zip_text_file(path(FOLDER, 'sample1.txt.gzip')))
        self.assertListEqual(array,Query.file.zipped_text(path(FOLDER,'sample1.txt.gzip')).to_list())





    def test_pickle_unpickle(self):
        os.makedirs(path(FOLDER),exist_ok=True)
        Query.en(range(10)).select(DTO).feed(fluq.to_pickle_file(path(FOLDER, 'sample.pickle')))
        result = Query.file.pickle(path(FOLDER,'sample.pickle')).with_indices().all(lambda z:
          z.item.num == z.index and
          z.item.s == 'test'*z.index and
          len(z.item.array) == z.index)
        self.assertTrue(result)



from tests.core.common import *
import os
import time
from yo_extensions.fluq_initiators import FileQuery
from yo_extensions.fluq import to_text_file

def _time_consuming_selector(q): # pragma: no cover
    time.sleep(0.1)
    return q



class ParallelTests(LinqTestBase):
    def test_fork(self):
        ctx = ForkContext(dict(factor=2, buffer = []))
        full_result = (
                Query
                .en(range(15))
                .fork(ctx, lambda query, context: Query
                      .en(query)
                      .where(lambda z: z%context['factor']==0)
                      .foreach_and_continue(lambda z: context['buffer'].append(z))
                      .select(lambda z: -z)
                      .to_list())
                .to_list()
        )
        self.assertListEqual(list(range(15)),full_result)
        self.assertListEqual([], ctx.sent['buffer'])
        self.assertListEqual([z for z in range(15) if z%2==0], ctx.received['buffer'])
        self.assertListEqual([-z for z in range(15) if z%2==0], ctx.result)


    def test_fire_and_forget(self):
        path = 'parallel.test.txt'
        if os.path.isfile(path):  # pragma: no cover
            os.remove(path)
        full_result = (Query
                       .en(range(15))
                       .fire_and_forget(lambda q: Query.en(q).where(lambda z: z%2==0).feed(to_text_file(self.path('test_fluq', path))))
                       .to_list()
                       )
        self.assertListEqual(list(range(15)),full_result)
        self.assertListEqual([z for z in range(15) if z%2==0], FileQuery().text(self.path('test_fluq',path)).select(int).to_list())

    def test_parallel_select_0(self):
        start_time = time.time()
        result = Query.en(range(10)).parallel_select(_time_consuming_selector,4,1).to_list()
        duration = time.time()-start_time
        self.assertLess(duration,0.5)




from tests.extensions.common import *

class FractionsTests(TestCase):
    def test_groupby(self):
        df = pd.DataFrame({'a':[1,1,2,2,2]})
        series = df.groupby('a').feed(alg.fractions()).feed(Query.series).to_dictionary()
        self.assertDictEqual({1:0.4,2:0.6},series)
    def test_series(self):
        s = pd.Series([1,2,2]).feed(alg.fractions()).feed(list)
        self.assertListEqual([0.2,0.4,0.4],s)

    def test_else(self):
        self.assertRaises(TypeError,lambda: alg.fractions()(1))


from tests.core.common import *


def _lazy():
    for i in range(10):
        yield i

class InitiatorsTest(LinqTestBase):
    def assertKeyValuePair(self, keys, values, results):
        for key, value, result in zip(keys,values,results):
            self.assertIsInstance(result,KeyValuePair)
            self.assertEqual(key,result.key)
            self.assertEqual(value,result.value)

    def test_sized_en(self):
        self.assertEqual(3,Query.en([1,2,3]).length)

    def test_unsized_en(self):
        self.assertEqual(None,Query.en(_lazy()).length)

    def test_unsized_count(self):
        self.assertEqual(10,Query.en(_lazy()).count())

    def test_sized_select(self):
        self.assertEqual(3,Query.args(1,2,3).select(str).length)

    def test_size_args(self):
        self.assertEqual(3,Query.args(1,2,3).length)


    def test_dict(self):
        self.assertKeyValuePair(
            ['a','b','c'],
            [1,2,3],
            Query.dict(Obj(a=1,b=2,c=3))
        )

    def test_series(self):
        self.assertKeyValuePair(
            ['a','b','c'],
            [1,2,3],
            Query.series(pd.Series([1,2,3],['a','b','c']))

        )

    def test_df(self):
        df = pd.DataFrame(columns=dict(testA=[1,2,3], testB=['1','2','3']))
        result = Query.df(df).with_indices().all(lambda z:
            z.item.testA == z.index and
            z.item.testB == str(z.index) and
            isinstance(z,Obj))
        self.assertTrue(result)



from typing import *
from yo_core import *
import gzip
import pickle

class to_text_file(Callable[[Queryable],None]):
    def __init__(self, filename: str, autoflush: bool = False, separator: str ='\n', **kwargs):
        self.filename = filename
        self.autoflush = autoflush
        self.separator = separator
        self.kwargs = kwargs

    def __call__(self, query: Queryable):
        with open(self.filename, 'w', **self.kwargs) as file:
            for item in query:
                file.write(str(item) + self.separator)
                if self.autoflush:
                    file.flush()

class to_zip_text_file(Callable[[Queryable],None]):
    def __init__(self, filename: str, separator: str = '\n'):
        self.filename = filename
        self.separator = separator

    def __call__(self, query):
        with gzip.open(self.filename, 'wb') as f:
            for item in query:
                item += self.separator
                item = bytes(item, 'utf-8')
                f.write(item)


class to_pickle_file(Callable[[Queryable],None]):
    def __init__ (self, filename: str):
        self.filename = filename

    def __call__(self, query):
        with open(self.filename,'wb') as file:
            for item in query:
                dump = pickle.dumps(item)
                length = len(dump)
                length_bytes = length.to_bytes(4,byteorder='big')
                file.write(length_bytes)
                file.write(dump)


import unittest
import os
from yo_core import *
from yo_extensions.misc import find_root_folder


class LinqTestBase(unittest.TestCase):

    def assertQuery(self, q, *args):
        self.assertListEqual(
            list(args),
            list(q)
        )

    def path(self, *args):
        folder = find_root_folder('yo.root')
        return os.path.join(folder,'tests','core',*args)






from yo_extensions import *
from yo_extensions.plots import *
from unittest import TestCase
from yo_extensions.misc import *

def path(*args):
    folder = find_root_folder('yo.root')
    return os.path.abspath(os.path.join(folder, 'tests', 'extensions', 'test_plots_extensions', *args))


import yaml
import pickle
import json
import os
import jsonpickle

class IO:
    @staticmethod
    def read_yaml(filename):
        with open(filename,'r') as file:
            return yaml.load(file)

    @staticmethod
    def read_pickle(filename):
        with open(filename,'rb') as file:
            return pickle.load(file)

    @staticmethod
    def read_json(filename):
        with open(filename,'r') as file:
            return json.load(file)


    @staticmethod
    def read_text(filename):
        with open(filename,'r') as file:
            return file.read()

    @staticmethod
    def read_jsonpickle(filename):
        with open(filename,'r') as file:
            return jsonpickle.loads(file.read())

    @staticmethod
    def write_yaml(data, filename):
        with open(filename,'w') as file:
            yaml.dump(data,file)

    @staticmethod
    def write_pickle(data, filename):
        with open(filename,'wb') as file:
            pickle.dump(data,file)

    @staticmethod
    def write_json(data, filename):
        with open(filename,'w') as file:
            json.dump(data,file,indent=1)

    @staticmethod
    def write_text(data, filename):
        with open(filename,'w') as file:
            file.write(data)

    @staticmethod
    def write_jsonpickle(data, filename):
        with open(filename,'w') as file:
            file.write(jsonpickle.dumps(data))

    @staticmethod
    def relative_to_file(file, *path):
        return os.path.join(os.path.dirname(file),*path)


    @staticmethod
    def find_root_folder(root_file_name):
        root_path = './'
        for i in range(100):
            if os.path.isfile(root_path + root_file_name):
                return root_path
            root_path += '../'
        raise ValueError(
            "Cound't find the root {1}. The current directory is {0}".format(os.path.abspath('.'), root_file_name))




from typing import *
import pandas as pd

class fractions(Callable[[Union[pd.core.groupby.DataFrameGroupBy,pd.Series]],pd.Series]):
    def __init__(self):
        pass

    def __call__(self,
                 data: Union[pd.core.groupby.DataFrameGroupBy,pd.Series]) -> pd.Series:
        if isinstance(data,pd.core.groupby.DataFrameGroupBy):
            size = data.size()
            return size/size.sum()
        elif isinstance(data,pd.Series):
            return data/data.sum()
        else:
            raise TypeError('Only Series, DataFrameGroupBy are supported')




import math
from matplotlib import pyplot as plt
import tqdm
from typing import *
from matplotlib.axes import Axes
from yo_core._fluq import *

T = TypeVar('T')

class ItemWithPlot(Generic[T]):
    def __init__(self, item: T, ax: Axes, all_axes = Any):
        self.item = item
        self.ax = ax
        self._all_axes = all_axes


class with_plots(Callable[[Queryable[T]],Queryable[ItemWithPlot[T]]]):
    def __init__(self,columns: int, figsize: Tuple[int,int] = (5,3), name_selector : Callable[[Any],str] = lambda x: str(x)):
        self.columns = columns
        self.figsize = figsize
        self.name_selector = name_selector

    def _make(self, q: Queryable):
        q = q.to_list()
        count = len(q)
        rows = int(math.ceil(count / self.columns))
        fig, axes = plt.subplots(rows, self.columns, figsize=(self.figsize[0] * self.columns, self.figsize[1] * rows))

        r_ax = [axes]
        if isinstance(axes,Collection):
            if isinstance(axes[0],Collection):
                r_ax = Query.en(axes).select_many(lambda z: z).to_list()
            else:
                r_ax = list(axes)

        for ax, element in zip(r_ax,q):
            if self.name_selector is not None:
                ax.set_title(self.name_selector(element))
            yield ItemWithPlot(element, ax, axes)

    def __call__(self, q: Queryable[T]) -> Queryable[ItemWithPlot[T]]:
        return Queryable(self._make(q),q.length)


def _isnotebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False


class with_progress_bar(Callable[[Queryable[T]],Queryable[T]]):
    def __init__(self, hide=False, console=None, **kwargs):
        if console is None:
            console = not _isnotebook()
        self.console = console
        self.hide = hide
        self.kwargs = kwargs

    def __call__(self, q: Queryable[T])->Queryable[T]:
        if self.hide:
            return q
        try:
            length = q.length
            if length is not None:
                self.kwargs['total'] = length
        except:
            pass
        if not self.console:
            return Queryable(tqdm.tqdm_notebook(q, **self.kwargs),q.length)
        else:
            return Queryable(tqdm.tqdm(q, **self.kwargs),q.length)




import unittest
from yo_core._common.ordered_enum import *
import pandas as pd

class TestEnum(OrderedEnum):
    LESS = 0
    GREATER = 1

class TestOrderedEnum(unittest.TestCase):
    def test_ordered_enum(self):
        self.assertGreater(TestEnum.GREATER,TestEnum.LESS)
        self.assertGreaterEqual(TestEnum.GREATER, TestEnum.LESS)
        self.assertLess(TestEnum.LESS,TestEnum.GREATER)
        self.assertLessEqual(TestEnum.LESS,TestEnum.GREATER)

        self.assertRaises(NotImplementedError, lambda:TestEnum.GREATER > 0)
        self.assertRaises(NotImplementedError, lambda:TestEnum.GREATER >= 0)
        self.assertRaises(NotImplementedError, lambda:TestEnum.GREATER < 0)
        self.assertRaises(NotImplementedError, lambda:TestEnum.GREATER <= 0)

    def test_ord_enum_in_pd(self):
        df = pd.DataFrame({'a':[TestEnum.GREATER,TestEnum.GREATER,TestEnum.LESS],'b':[1,1,2]})
        df.groupby('a').size()



from yo_extensions.misc import *
from tests.extensions.common import *
from IPython.display import HTML

class MiscTests(TestCase):
    def test_json(self):
        os.makedirs(path('misc_files'),exist_ok=True)
        jsonpath = path('misc_files','test.json')
        obj = dict(a=1,b='2')
        save_json(jsonpath,obj)
        self.assertDictEqual(
            dict(a=1,b='2'),
            load_json(jsonpath))

    def test_json_1(self):
        jsonpath = path('misc_files', 'test1.json')
        obj = dict(a=1, b='2')
        save_json(jsonpath, obj)
        self.assertDictEqual(
            dict(a=1, b='2'),
            load_json(jsonpath,as_obj=True))


    def test_pickle(self):
        jsonpath = path('misc_files', 'test.pkl')
        obj = dict(a=1, b='2')
        save_pkl(jsonpath, obj)
        self.assertDictEqual(
            dict(a=1, b='2'),
            load_pkl(jsonpath))

    def test_printable(self):
        self.assertIsInstance(notebook_printable_version(True),HTML)
        self.assertIsNone(notebook_printable_version(False))

    def test_find_root_fails(self):
        self.assertRaises(ValueError,lambda: find_root_folder('aigjixfjbvijfdawemnarldsncolkcznvoisalmenfwsefdfsdvlskdapejr'))


    def test_diffset_no(self):
        r = diffset([1,2],[2,3])
        self.assertFalse(r.Match)
        self.assertFalse(r.LeftIsSubset)
        self.assertFalse(r.RightIsSubset)
        self.assertEqual(1,r.Intersection)

    def test_diffset_left(self):
        r = diffset([1,2],[1,2,3])
        self.assertFalse(r.Match)
        self.assertTrue(r.LeftIsSubset)
        self.assertFalse(r.RightIsSubset)
        self.assertEqual(2,r.Intersection)


    def test_diffset_right(self):
        r = diffset([1,2, 3],[2,3])
        self.assertFalse(r.Match)
        self.assertFalse(r.LeftIsSubset)
        self.assertTrue(r.RightIsSubset)
        self.assertEqual(2,r.Intersection)


    def test_diffset_match(self):
        r = diffset([1,2, 3],[1, 2,3])
        self.assertTrue(r.Match)
        self.assertTrue(r.LeftIsSubset)
        self.assertTrue(r.RightIsSubset)
        self.assertEqual(3,r.Intersection)



from tests.extensions.common import *
from tqdm import tqdm
from matplotlib.axes import Axes

class YoWithTests(TestCase):
    def test_progress_bar_0(self):
        en = Query.args(1,2,3).feed(fluq.with_progress_bar(console=True)).en
        self.assertTrue(isinstance(en,tqdm))
        self.assertEqual(3,en.total)

    def test_progress_bar_1(self):
        en = Query.args(1,2,3).where(lambda z: True).feed(fluq.with_progress_bar(console=True)).en
        self.assertTrue(isinstance(en,tqdm))
        self.assertEqual(None,en.total)

    def test_plots_0(self):
        result = Query.args("1","2","3").feed(fluq.with_plots(2)).to_list()
        axes = result[0]._all_axes
        self.assertEqual((2,2),axes.shape)
        self.assertEqual('1', axes[0, 0].title.get_text())
        self.assertEqual('2', axes[0, 1].title.get_text())
        self.assertEqual('3', axes[1, 0].title.get_text())
        self.assertEqual('1',result[0].item)


    def test_plots_1(self):
        result = Query.args('1').feed(fluq.with_plots(1)).to_list()
        axes = result[0]._all_axes
        self.assertTrue(isinstance(axes,Axes))





from tests.extensions.common import *
import numpy as np

df = Query.en(['A']*30+['B'] * 60+['C'] * 10).to_dataframe(columns=['Group']).reset_index()

class StratifyColumnTests(TestCase):

    def chech(self,res):
        self.assertListEqual(list(range(100)), list(res.index.sort_values()))
        due = Query.series(df.groupby('Group').feed(alg.fractions())).to_dictionary()
        width = 15
        for i in range(df.shape[0] - width):
            windowed = res.iloc[i:i + width].groupby('Group').feed(alg.fractions())
            for key, value in due.items():
                self.assertLess(abs(value - windowed[key]), 0.1)

    def test_1(self):
        res = df.feed(alg.stratify('Group','stratify'))
        self.chech(res)

    def test_2(self):
        res = df.feed(alg.stratify('Group','stratify',random_state=42))
        self.chech(res)

    def test_3(self):
        res = df.feed(alg.stratify('Group','stratify',random_state=np.random.RandomState(45)))
        self.chech(res)




from .fluq_initiators import FileQuery as _FileQuery, CombinatoricsQuery as _CombinatoricsQuery, folder as _folder, LoopEndType, loop_maker as _loop_maker
from yo_core import *
from yo_core import Query as QueryBase
from pathlib import Path

from . import alg, fluq, plots

class Query(QueryBase):
    file = _FileQuery()
    combinatorics = _CombinatoricsQuery()
    @staticmethod
    def folder(location: Union[Path, str], pattern: str = '*') -> Queryable[Path]:
        return Queryable(_folder(location, pattern))
    @staticmethod
    def loop(begin: Any, delta: Any, end: Any = None, endtype=LoopEndType.NotEqual):
        lp = _loop_maker(begin,delta,end,endtype)
        return Queryable(lp.make())



import pandas as pd
from typing import *

T = TypeVar('T')
TOut = TypeVar('TOut')

def _feed(object: T, method: Callable[[T],TOut])->TOut:
    return method(object)


pd.Series.feed = _feed
pd.DataFrame.feed = _feed
pd.core.groupby.DataFrameGroupBy.feed = _feed
pd.core.groupby.SeriesGroupBy.feed = _feed



from .generics import *
from .agg_methods import _or_default


class ItemWithIndex(Generic[T]):
    def __init__(self, index: int, item: T):
        self.index = index
        self.item = item

def with_indices(en):
    for index,item in enumerate(en):
        yield ItemWithIndex(index,item)


def foreach(en, action):
    for element in en:
        if action is not None:
            action(element)

def foreach_and_continue(en,action):
    for element in en:
        action(element)
        yield element

def arg_extremum_default(en, selector, is_max, with_default, default):
    argmax = None
    argmax_item = None
    first_time = True
    for item in en:
        if first_time:
            first_time = False
            argmax_item=item
            argmax = selector(item)
        else:
            value = selector(item)
            if (is_max and value>argmax) or (not is_max and value<argmax):
                argmax = value
                argmax_item = item
    if not first_time:
        return argmax_item

    return _or_default(with_default, default)

def append(en,array):
    for e in en:
        yield e
    for e in array:
        yield e

def prepend(en, array):
    for e in array:
        yield e
    for e in en:
        yield e





import pandas as pd
import numpy as np
from .add_ordering_column import add_ordering_column
from typing import *


#https://cs.stackexchange.com/questions/29709/algorithm-to-distribute-items-evenly
#by https://cs.stackexchange.com/users/70760/lungj
def _generate(item_counts):
    '''
    item_counts is a list of counts of "types" of items. E.g., [3, 1, 0, 2] represents
    a list containing [1, 1, 1, 2, 4, 4] (3 types of items/distinct values). Generate
    a new list with evenly spaced values.
    '''
    # Sort number of occurrences by decreasing value.
    item_counts.sort(reverse=True)
    # Count the total elements in the final list.
    unplaced = sum(item_counts)
    # Create the final list.
    placements = [None] * unplaced

    # For each type of item, place it into the list item_count times.
    for item_type, item_count in enumerate(item_counts):
        # The number of times the item has already been placed
        instance = 0
        # Evenly divide the item amongst the remaining unused spaces, starting with
        # the first unused space encountered.
        # blank_count is the number of unused spaces seen so far and is reset for each
        # item type.
        blank_count = -1
        for position in range(len(placements)):
            if placements[position] is None:
                blank_count += 1
                # Use an anti-aliasing technique to prevent bunching of values.
                if blank_count * item_count // unplaced == instance:
                    placements[position] = item_type
                    instance += 1
        # Update the count of number of unplaced items.
        unplaced -= item_count

    return placements

def _generate_index(placements, counts):
    result = []
    for _class in placements:
        counts[_class]-=1
        result.append(counts[_class])
    return result

class stratify(Callable[[pd.DataFrame],pd.DataFrame]):
    def __init__(self,
        columns: Union[str,List[str]],
        stratify_column_name: str='stratify_index',
        random_state: Optional[Union[int,np.random.RandomState]]= None):
        self.columns = columns
        self.stratify_column_name = stratify_column_name
        self.random_state = random_state

    def __call__(self, df: pd.DataFrame)->pd.DataFrame:
        if self.random_state is None:
            self.random_state = np.random.RandomState()
        elif isinstance(self.random_state,int):
            self.random_state = np.random.RandomState(self.random_state)

        random_order = self.random_state.permutation(df.shape[0])
        df = df.assign(_rand_order = random_order)
        df = add_ordering_column(self.columns, '_rand_order', '_order', '_group')(df)
        #print(df)

        counts = df.groupby('_group').size().to_frame('counts').sort_values('counts',ascending=False).reset_index()
        #print(counts)

        group_index = _generate(list(counts.counts))

        order_index = _generate_index(group_index,list(counts.counts))

        result = pd.DataFrame(dict(group_index=group_index, _order=order_index))
        result = result.merge(counts._group.to_frame(),left_on='group_index',right_index=True).sort_index()
        result = result.reset_index().rename(columns=dict(index=self.stratify_column_name)).drop('group_index',axis=1)
        result = result.set_index(['_group','_order'])
        result = result.merge(df,left_index=True,right_on=['_group','_order']).set_index(self.stratify_column_name)
        result = result.drop(['_rand_order','_order','_group'],axis=1)
        return result





from .core_linq import *
from .original_asq_ordering import *
from .agg_methods import *
from .extended_methods import *
from .parallel import *
from .generics import KeyValuePair
from .aggregation_methods import *


from typing import *
import pandas as pd
import numpy as np

class _SideTrimmer:
    def __init__(self, mode, value, percentile):
        self.mode = mode
        self.value = value
        self.percentile = percentile

    def process(self, series, before):
        value = self.value
        data = np.array(series)
        if self.percentile is not None:
            value = np.percentile(data, self.percentile)
        if before:
            pattern = series > value
        else:
            pattern = series < value
        if self.mode == 'drop':
            return series[pattern]
        else:
            return series.where(pattern, value)


class trimmer(Callable[[pd.Series], pd.Series]):
    def __init__(self,
                 drop_before: Optional[float]=None,
                 drop_before_percentile: Optional[float]=None,
                 unite_before: Optional[float] =None,
                 unite_before_percentile: Optional[float] =None,
                 drop_after: Optional[float] =None,
                 drop_after_percentile: Optional[float] =None,
                 unite_after: Optional[float] =None,
                 unite_after_percentile: Optional[float] =None):
        self.before = None
        self.after = None

        if drop_before is not None:
            self.before = _SideTrimmer('drop', drop_before, None)
        if drop_before_percentile is not None:
            self.before = _SideTrimmer('drop', None, drop_before_percentile)
        if unite_before is not None:
            self.before = _SideTrimmer('unite', unite_before, None)
        if unite_before_percentile is not None:
            self.before = _SideTrimmer('unite', None, unite_before_percentile)

        if drop_after is not None:
            self.after = _SideTrimmer('drop', drop_after, None)
        if drop_after_percentile is not None:
            self.after = _SideTrimmer('drop', None, drop_after_percentile)
        if unite_after is not None:
            self.after = _SideTrimmer('unite', unite_after, None)
        if unite_after_percentile is not None:
            self.after = _SideTrimmer('unite', None, unite_after_percentile)

    def __call__(self, series: pd.Series) ->pd.Series:
        if self.before is not None:
            series = self.before.process(series, True)
        if self.after is not None:
            series = self.after.process(series, False)
        return series



from tests.core.common import *


class GeneralSelectorsTests(LinqTestBase):

    def test_select(self):
        self.assertQuery(Query.args(1, 2).select(str), '1', '2')

    def test_select_many(self):
        self.assertQuery(Query.args(1, 2, 3).select_many(lambda z: range(z)), 0, 0, 1, 0, 1, 2)

    def test_distinct(self):
        self.assertQuery(Query.args(1, 2, 2, 3, 3).distinct(), 1, 2, 3)

    def test_distinct_1(self):
        self.assertQuery(Query.args('aa', 'ab', 'b').distinct(lambda z: z[0]), 'aa', 'b')

    def test_where(self):
        self.assertQuery(Query.args(1, 2, 3, 4).where(lambda z: z % 2 == 0), 2, 4)

    def test_group_by(self):
        result = Query.args(2, 3, 1, 4).group_by(lambda z: z % 2).to_list()
        self.assertEqual(0, result[0].key)
        self.assertListEqual([2, 4], result[0].value)
        self.assertEqual(1, result[1].key)
        self.assertListEqual([3, 1], result[1].value)

    def test_group_by_1(self):
        result = Query.args(2,3,1,4).group_by(lambda z: z%2).select(lambda z: (z.key,Query.en(z).sum())).to_list()
        self.assertListEqual(
            [(0,6),(1,4)],
            result)

    def test_group_by_2(self):
        result = Query.args(2,3,1,4).group_by(lambda z: z%2).to_dictionary()
        self.assertDictEqual({0:[2,4],1:[3,1]},result)

    def test_aggrerate(self):
        self.assertEqual('123', Query.args('1', '2', '3').aggregate(lambda acc, el: acc + el))

    def test_cast_1(self):
        self.assertRaises(TypeError,lambda: Query.args(1,2,"3").cast(int).to_list())

    def test_cast_2(self):
        self.assertQuery(Query.args(1, 2, 3).cast(int), 1, 2, 3)

    def test_of_type(self):
        self.assertQuery(
            Query.args(1,"2",3,"4",5).of_type(int),
            1,3,5)

    def test_concat(self):
        self.assertQuery(
            Query.args(1,2,3).concat([3,4,5]),
            1,2,3,3,4,5
        )

    def test_intersect(self):
        self.assertQuery(
            Query.args(1,2,3).intersect([2,3,4]),
            2,3
        )




from sklearn import datasets
from yo_ml.keras import *

import unittest
from tests.ml.common import *
from tests.ml.common import path as find_path

feature_columns = ['F1', 'F2', 'F3', 'F4']
classes_count = 3
path = find_path('keras')

def make_df() -> pd.DataFrame:
    iris = datasets.load_iris()
    df = pd.DataFrame(iris['data'], columns=feature_columns)
    df = df.assign(target=iris['target'])
    df = df.feed(alg.stratify('target', random_state=42))
    df = df.assign(in_train=np.where(df.index < df.shape[0] * 3 / 4, True, False))
    return df

class ModelFactory(ModelFactoryBase):
    def __init__(self, df):
        self.df = df

    def mapper_x(self, ids):
        return self.df.loc[ids][feature_columns].values

    def mapper_y(self, ids):
        result = np.zeros((len(ids),classes_count))
        for position_1, id in enumerate(ids):
            result[position_1,self.df.loc[id].target]=1
        return result

    def create_x_featurizer(self):
        return Featurizer(self.mapper_x)

    def create_y_featurizer(self):
        return Featurizer(self.mapper_y)

    def create_model(self):
        model = keras.models.Sequential()
        model.add(keras.layers.Dense(
            8,
            input_shape=(len(feature_columns),),
            activation='relu',
            name='hidden'))
        model.add(keras.layers.Dense(
            classes_count,
            activation='softmax',
            name='output'))
        model.compile(loss=keras.losses.categorical_crossentropy, optimizer='adam')
        return model

def run_train(df, folder, call_count, epochs, reset):
    process = KerasTrainTest(ModelFactory(df), os.path.join(path, folder), reset=reset)
    process.make_all_train(
        list(df.loc[df.in_train].index),
        call_count,
        epochs=epochs,
        batch_size=5,
        verbose=1
    )



class KerasTests(unittest.TestCase):
    def evaluate(self, df, folder):
        process = KerasTrainTest(ModelFactory(df), os.path.join(path, folder))
        result = (process
                  .evaluate(list(df.index))
                  .select(lambda z: (df.iloc[z.id].in_train, z.output.argmax(), z.answer.argmax()))
                  .to_dataframe(columns=['in_train', 'predicted', 'actual'])
                  )
        wrongs = (result.loc[result.predicted != result.actual]).shape[0]
        #self.assertGreater(7,wrongs)

    def test_simple(self):
        df = make_df()
        epochs = 5
        run_train(df,'simple',1,epochs, True)
        history = load_pkl(os.path.join(path,'simple','history'))
        self.assertEqual(epochs,len(history[0]['loss']))
        self.evaluate(df,'simple')






from tests.extensions.common import *

class ExtendedMethodsTests(TestCase):
    def test_count_partition(self):
        lsts = Query.args(1,2,3,4,5).feed(fluq.partition_by_count(2)).to_list()
        self.assertEqual(3,len(lsts))
        self.assertListEqual([1,2],lsts[0])
        self.assertListEqual([3,4],lsts[1])
        self.assertListEqual([5],lsts[2])

    def test_count_empty_partition(self):
        lsts = Query.args().feed(fluq.partition_by_count(2)).to_list()
        self.assertEqual(0,len(lsts))

    def test_selector_partition(self):
        lsts = (Query
                .args( ('a',1), ('a',2), ('b',3), ('c',4), ('c', 5))
                .feed(fluq.partition_by_selector(lambda z: z[0]))
                .select(lambda z: [x[1] for x in z])
                .to_list()
                )
        self.assertEqual(3,len(lsts))
        self.assertListEqual([1,2],lsts[0])
        self.assertListEqual([3], lsts[1])
        self.assertListEqual([4,5],lsts[2])


from .builder_helpers import *
import itertools



class Queryable(Generic[T]):
    def __init__(self, en: Iterable[T], length: Optional[int] = None):
        self.en=en
        self.length = length

    def __iter__(self):
        for e in self.en:
            yield e

    # <editor-fold desc="core_linq">
    def select(self, selector: Callable[[T],TOut]) -> 'Queryable[TOut]':
        return Queryable(map(selector, self.en), self.length)

    def distinct(self, selector: Optional[Callable[[T],TOut]] = None) -> 'Queryable[T]':
        return Queryable(distinct(self.en, selector))

    def select_many(self, selector: Callable[[T],Iterable[TOut]]) -> 'Queryable[TOut]':
        return Queryable(itertools.chain.from_iterable(map(selector, self.en)))

    def where(self, filterSelector: Callable[[T],bool]) -> 'Queryable[T]':
        return Queryable(filter(filterSelector, self.en))

    def group_by(self, selector: Callable[[T],TKey])-> 'Queryable[Group[TKey,T]]':
        return Queryable(group_by(self.en, selector))

    def cast(self, _type: Type[TOut]) -> 'Queryable[TOut]':
        return Queryable(fluq_cast(self.en, _type), self.length)

    def of_type(self, _type: Type[TOut]) -> 'Queryable[TOut]':
        return Queryable(of_type(self.en, _type), None)

    def intersect(self, en2: Iterable[T])->'Queryable[T]':
        return Queryable(intersect(self.en,en2))

    def concat(self, en2: Iterable[T])->'Queryable[T]':
        return Queryable(concat(self.en, en2))

    # </editor-fold>

    # <editor-fold desc="original_asq_ordering">
    def order_by(self, selector: Callable[[T], Any]) -> 'Queryable[T]':
        return Queryable(Orderer(self.en, [(-1, selector)]), self.length)

    def order_by_descending(self, selector: Callable[[T], Any]) -> 'Queryable[T]':
        return Queryable(Orderer(self.en, [(1, selector)]), self.length)

    def then_by(self, selector: Callable[[T], Any]) -> 'Queryable[T]':
        if not isinstance(self.en, Orderer):
            raise ValueError('then_by can only be called directly after order_by or order_by_descending')
        return Queryable(Orderer(self.en, self.en._funcs + [(-1, selector)]), self.length)

    def then_by_descending(self, selector: Callable[[T], Any]) -> 'Queryable[T]':
        if not isinstance(self.en, Orderer):
            raise ValueError('then_by can only be called directly after order_by or order_by_descending')
        return Queryable(Orderer(self.en, self.en._funcs + [(1, selector)]), self.length)
    # </editor-fold>

    # <editor-fold desc="various methods">

    def aggregate(self, aggregator: Callable[[T,T],T])->T:
        return aggregate(self.en,aggregator)

    def aggregate_with(self, *args: BaseAggregation):
        return aggregate_with(self.en, None, *args)

    def count(self) -> int:
        return aggregate_with(self.en, None, Count())

    def sum(self, selector: Optional[Callable[[T],Any]] = None) -> Any:
        return aggregate_with(self.en, selector, Sum())

    def all(self, selector: Callable[[T],bool] = lambda z: True)->bool:
        return aggregate_with( self.en, selector, All())

    def any(self, selector: Callable[[T],bool] = lambda z: True)->bool:
        return aggregate_with( self.en, selector, Any() )

    def mean(self, selector: Optional[Callable[[T],float]] = None) -> float:
        return aggregate_with(self.en, selector, Mean() )

    def min(self, selector: Optional[Callable[[T], Any]] = None) -> Any:
        return aggregate_with(self.en, selector, Min())

    def max(self, selector: Optional[Callable[[T], Any]] = None) -> Any:
        return aggregate_with(self.en, selector, Max())

    # </editor-fold>

    # <editor-fold desc="various methods">
    def skip(self, count) -> 'Queryable[T]':
        return Queryable(itertools.islice(self.en, count, None))

    def take(self, count) -> 'Queryable[T]':
        return Queryable(itertools.islice(self.en, count))

    def skip_while(self, condition) -> 'Queryable[T]':
        return Queryable(skip_while(self.en, condition))

    def take_while(self, condition) -> 'Queryable[T]':
        return Queryable(take_while(self.en, condition))

    def first(self)->T:
        return first(self.en, False)

    def first_or_default(self, default = None) -> T:
        return first(self.en, True, default)

    def last(self) -> T:
        return last(self.en, False)

    def last_or_default(self, default=None) -> T:
        return last(self.en, True, default)

    def single(self) -> T:
        return single(self.en, False)

    def single_or_default(self, default=None)->T:
        return single(self.en, True, default)


    def to_list(self)->List[T]:
        return list(self.en)

    def to_dictionary(self,key_selector: Optional[Callable[[T],TKey]]=None, value_selector: Optional[Callable[[T],TValue]]=None) -> Dict[TKey,TValue]:
        return to_dictionary(self.en,key_selector,value_selector)

    def to_tuple(self)->Tuple:
        return tuple(self.en)

    def to_set(self)->Set[T]:
        return to_set(self.en)


    def to_ndarray(self) -> np.ndarray:
        return to_ndarray(self.en)

    def to_series(self,
               item_to_value: Optional[Callable[[T], Any]] = None,
               item_to_index: Optional[Callable[[T], Any]] = None) -> pd.Series:
        return to_series(self.en, item_to_value, item_to_index)

    def to_dataframe(self, **kwargs) -> pd.DataFrame:
        return to_dataframe(self.en, **kwargs)

    # </editor-fold>

    # <editor-fold desc="extended_methods">
    def argmax(self, selector: Callable[[T],Any])->T:
        return arg_extremum_default(self.en,selector,True,False,None)

    def argmax_or_default(self, selector: Callable[[T],Any], default=None):
        return arg_extremum_default(self.en,selector,True,True,default)

    def argmin(self,selector: Callable[[T],Any])->T:
        return arg_extremum_default(self.en,selector,False,False,None)

    def argmin_or_default(self, selector: Callable[[T],Any], default=None):
        return arg_extremum_default(self.en, selector, False, True, default)

    def with_indices(self)-> 'Queryable[ItemWithIndex[T]]':
        return Queryable(with_indices(self.en), self.length)

    def foreach(self, action: Callable[[T],None])->None:
        foreach(self.en,action)

    def foreach_and_continue(self, action: Callable[[T],None])-> 'Queryable[T]':
        return Queryable(foreach_and_continue(self.en, action), self.length)

    def append(self,*args:T)-> 'Queryable[T]':
        return Queryable(append(self.en, args))

    def prepend(self, *args:T)-> 'Queryable[T]':
        return Queryable(prepend(self.en, args))
    # </editor-fold>


    #<editor-fold desc="Parallel">

    def fork(self, context: ForkContext, pipeline: Callable[[Any,Iterable[T]],Any]):
        return Queryable(fork(self.en,context, pipeline))

    def fire_and_forget(self, pipeline: Callable[[Iterable[T]],Any]):
        return Queryable(fork(self.en, ForkContext(None), lambda q, _ : pipeline(q)))

    def parallel_select(self, selector, workers_count=None, buffer_size=1):
        return Queryable(parallel_select(self.en, selector, workers_count, buffer_size),self.length)

    #</editor>

    #<editor-fold desc="Extension Mechanism">

    def feed(self, collector: Callable[['Queryable[T]'], TOut])->TOut:
        return collector(self)

    #</editor-fold>


from yo_extensions import *
from yo_extensions.misc import *
import keras
import numpy as np
from keras.utils import plot_model


class DataGenerator(keras.utils.Sequence):
    'Generates data for Keras'
    def __init__(
            self,
            sample_ids,
            input_selector,
            output_selector,
            batch_size=64,
            shuffle=True,
            sample_weights_selector = None
    ):
        self.sample_ids = sample_ids
        self.input_selector = input_selector
        self.output_selector = output_selector
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.samples_weights_selector = sample_weights_selector
        self.on_epoch_end()

    def __len__(self):
        'Denotes the number of batches per epoch'
        return int(np.ceil(len(self.sample_ids) / self.batch_size))

    def on_epoch_end(self):
        'Updates indexes after each epoch'
        self.indexes = np.arange(len(self.sample_ids))
        if self.shuffle == True:
            np.random.shuffle(self.indexes)

    def __getitem__(self, index):
        'Generate one batch of data'
        indexes = self.indexes[index * self.batch_size:(index + 1) * self.batch_size]
        samples = [self.sample_ids[k] for k in indexes]
        result = self.get_samples(samples)
        return result

    def get_samples(self, ids):
        Xs = self.input_selector(ids)
        ys = self.output_selector(ids)
        if self.samples_weights_selector is not None:
            ws = np.array([self.samples_weights_selector(id) for id in ids])
            return Xs, ys, ws
        else:
            return Xs, ys


class Featurizer:
    def __init__(self, *args: Callable, **kwargs: Callable):
        if (len(args)>0 and len(kwargs)>0) or (len(args)==0 and len(kwargs)==0):
            raise ValueError('Featurizer can accept exactly one of: args or kwargs')

        self.args = args
        self.kwargs = kwargs

    def get_features(self, arg):
        if len(self.args)==1:
            return self.args[0](arg)
        elif len(self.args)!=0:
            return [f(arg) for f in self.args]
        else:
            return {key:f(arg) for key, f in self.kwargs.items()}

    def get_feature_at_index(self, features, index):
        if len(self.args)==1:
            return features[index]
        elif len(self.args)!=0:
            return [f[index] for f in features]
        else:
            return {key: f[index] for key, f in features.items()}


class ModelFactoryBase:
    def create_model(self):
        raise NotImplementedError()
    def create_x_featurizer(self) -> Featurizer:
        raise NotImplementedError()
    def create_y_featurizer(self) -> Featurizer:
        raise NotImplementedError()
    def create_weight_selector(self) -> Optional[Callable[[Any],float]]:
        return None
    def get_settings(self):
        return {}


class EvaluationResult:
    def __init__(self, id: Any, input: np.ndarray, output: np.ndarray, answer: np.ndarray):
        self.id = id
        self.input = input
        self.output = output
        self.answer = answer

class KerasTrainTest:
    def __init__(self, factory: ModelFactoryBase, path: str, reset: bool = False):
        self.X_featurizer = factory.create_x_featurizer()
        self.Y_featurizer = factory.create_y_featurizer()
        self.path = path
        self.model_file = os.path.join(self.path,'model')
        self.history_file = os.path.join(self.path,'history')
        self.settings_file = os.path.join(self.path,'settings.json')
        self.weights_selector = factory.create_weight_selector()

        if os.path.isfile(self.model_file) and not reset:
            self.model = keras.models.load_model(self.model_file)
            self.histories = load_pkl(self.history_file)
        else:
            os.makedirs(self.path, exist_ok=True)
            self.model = factory.create_model()
            self.histories = []
            save_json(self.settings_file,factory.get_settings())
            plot_model(self.model, to_file=os.path.join(self.path, 'model.png'), show_shapes=True)
            self.save()

    def train(self, train_indices, validate_indices=None, batch_size=64, **kwargs):
        generator = DataGenerator(
            train_indices,
            self.X_featurizer.get_features,
            self.Y_featurizer.get_features,
            sample_weights_selector=self.weights_selector,
            batch_size = batch_size
        )

        if validate_indices is not None:
            validator = DataGenerator(
                validate_indices,
                self.X_featurizer.get_features,
                self.Y_featurizer.get_features
            )
        else:
            validator = None

        history = self.model.fit_generator(generator, validation_data=validator, **kwargs)
        self.histories.append(history.history)


    def save(self):
        try:
            self.model.save(self.model_file)
            save_pkl(self.history_file,self.histories)
        except:
            self.model.save(self.model_file) #this is mainly to resolve the problem of the corrupted file when KeyboardInterrupt
            save_pkl(self.history_file, self.histories)
            raise

    def make_all_train(self, train_indices,  call_count=1, **train_kwargs):
        for i in range(call_count):
            self.train(train_indices=train_indices, **train_kwargs)
            self.save()

    def _evaluate_iter(self, index, batch_size=64):
        batches = Query.en(index).feed(fluq.partition_by_count(batch_size)).to_list()
        q = Query.en(batches)
        for batch in q:
            input = self.X_featurizer.get_features(batch)
            output = self.model.predict_on_batch(input)
            answer = self.Y_featurizer.get_features(batch)
            for i in range(len(batch)):
                result = EvaluationResult(
                    batch[i],
                    self.X_featurizer.get_feature_at_index(input, i),
                    self.Y_featurizer.get_feature_at_index(output, i),
                    self.Y_featurizer.get_feature_at_index(answer, i))
                yield result

    def evaluate(self, index, batch_size=64)-> Queryable[EvaluationResult]:
        return Queryable(self._evaluate_iter(index,batch_size),len(index))




from tests.core.common import *

class OrderingTests(LinqTestBase):

    def test_order_by(self):
        self.assertQuery(Query.args('ccc', 'bb', 'a').order_by(len), 'a', 'bb', 'ccc')


    def test_order_by_descending(self):
        self.assertQuery(Query.args('ccc', 'bb', 'a').order_by_descending(len), 'ccc', 'bb', 'a')

    def test_order_then_1(self):
        self.assertQuery(Query.args('a2', 'b1', 'b2', 'a1').order_by(lambda z: z[0]).then_by(lambda z: z[1]), 'a1',
                   'a2', 'b1', 'b2')

    def test_order_then_2(self):
        self.assertQuery(
            Query.args('a2', 'b1', 'b2', 'a1').order_by(lambda z: z[0]).then_by_descending(lambda z: z[1]),
            'a2', 'a1', 'b2', 'b1')

    def test_order_then_3(self):
        self.assertQuery(
            Query.args('a2', 'b1', 'b2', 'a1').order_by_descending(lambda z: z[0]).then_by(lambda z: z[1]),
            'b1', 'b2', 'a1', 'a2')

    def test_order_then_4(self):
        self.assertQuery(
            Query.args('a2', 'b1', 'b2', 'a1').order_by_descending(lambda z: z[0]).then_by_descending(lambda z: z[1]),
            'b2', 'b1', 'a2', 'a1')

    def test_then(self):
        self.assertRaises(ValueError, lambda: Query.args(1, 2).then_by(lambda z: z).to_list())

    def test_then_desc(self):
        self.assertRaises(ValueError, lambda: Query.args(1, 2).then_by_descending(lambda z: z).to_list())




from tests.core.common import *
import pandas as pd
import numpy as np

class ToCollectionMethodsTests(LinqTestBase):
    def test_to_dictionary_normal(self):
        d = Query.args(1,2,3).to_dictionary(lambda z: z, lambda z: str(z))
        self.assertDictEqual({1:'1',2:'2',3:'3'},d)

    def test_to_dictionary_duplicating(self):
        self.assertRaises(ValueError,lambda: Query.args(1,1,2).to_dictionary(lambda z: z, lambda z: str(z)))

    def test_to_dictionary_key_value_pair(self):
        d = Query.args(1,2,3).select(lambda z: KeyValuePair(z,str(z))).to_dictionary()
        self.assertDictEqual({1: '1', 2: '2', 3: '3'}, d)

    def test_to_dictionary_key_value_pair_override(self):
        d = Query.dict(dict(a=1,b=2)).to_dictionary(lambda z: z.key.upper(), lambda z: z.value+1)
        self.assertDictEqual(dict(A=2,B=3),d)

    def test_to_dictionary_mixture(self):
        self.assertRaises(ValueError,lambda: Query.args(KeyValuePair(1,2),1).to_dictionary())

    def test_to_dictionary_without_key_selector(self):
        self.assertRaises(ValueError, lambda: Query.args(1).to_dictionary())

    def test_to_dictionary_without_value_selector(self):
        self.assertRaises(ValueError, lambda: Query.args(1).to_dictionary(lambda z: z))

    def test_to_dictionary_grby(self):
        result = Query.args(1,2,3,4).group_by(lambda z: z%2).to_dictionary()
        self.assertDictEqual({0:[2,4],1:[1,3]},result)

    def test_to_tuple(self):
        t = Query.args(1,2,3).to_tuple()
        self.assertIsInstance(t,tuple)
        self.assertListEqual([1,2,3],list(t))

    def test_to_set_1(self):
        st = Query.args(1,2,3).to_set()
        self.assertSetEqual({1,2,3},st)

    def test_to_set_2(self):
        self.assertRaises(ValueError,lambda: Query.args(1,2,2).to_set())

    def test_to_ndarray_1d(self):
        arr = Query.args(1,2,3).to_ndarray()
        self.assertEqual(type(arr),np.ndarray)

    def test_to_ndarray_2d(self):
        arr = Query.args([1,2],[3,4]).to_ndarray()
        self.assertListEqual([2,2],list(arr.shape))
        self.assertEqual(3,arr[1,0])

    def test_to_series_default(self):
        ser = Query.args(1,2,3).to_series()
        self.assertEqual(type(ser), pd.Series)
        self.assertListEqual([1,2,3],list(ser))

    def test_to_series_value_selector(self):
        ser = Query.args(1,2,3).to_series(str)
        self.assertListEqual(['1','2','3'],list(ser))

    def test_to_series_index_selector(self):
        ser = Query.args(1,2,3).to_series(item_to_index=str)
        self.assertListEqual([1,2,3],list(ser))
        self.assertListEqual(['1','2','3'], list(ser.index))

    def test_to_series_both_selectors(self):
        ser = Query.args(1, 2, 3).to_series(lambda z: -z, str)
        self.assertListEqual([-1, -2, -3], list(ser))
        self.assertListEqual(['1', '2', '3'], list(ser.index))

    def test_to_series_key_value_pair(self):
        ser = Query.args(1,2,3).select(lambda z: KeyValuePair(z,str(z))).to_series()
        self.assertListEqual([1,2,3],list(ser.index))
        self.assertListEqual(['1','2','3'], list(ser))

    def test_to_series_special_case(self):
        ser = Query.args(1,2,3,4).group_by(lambda z: z%2).to_series(lambda z: len(z),lambda z: z.key)
        self.assertListEqual([2,2],list(ser))
        self.assertEqual([1,0],list(ser.index))

    def test_to_df(self):
        data = Query.en([1,2,3]).select_many(lambda a: Query.en(['a','b']).select(lambda b: Obj(A=a,B=b))).to_dataframe()
        self.assertListEqual(['A','B'], list(data.columns))
        self.assertEqual('int64',data.A.dtype.name)
        self.assertEqual('object',data.B.dtype.name)
        self.assertListEqual([1, 1, 2, 2, 3, 3], list(data.A))
        self.assertListEqual(['a', 'b', 'a', 'b', 'a', 'b'], list(data.B))

    def test_to_df_variable_length(self):
        df = Query.args(Obj(a=1),Obj(a=1,b=2)).to_dataframe()
        self.assertListEqual(['a','b'],list(df.columns))



import unittest
from yo_core._common.obj import Obj, _EXCEPTIONS
import json

class ObjTests(unittest.TestCase):

    def test_init_and_getattr(self):
        obj = Obj(a=1)
        self.assertEqual(1,obj.a)

    def test_getattr_and_setattr(self):
        obj = Obj(a=1)
        obj.a=2
        self.assertEqual(2,obj.a)

    def test_setindex_and_getattr(self):
        obj = Obj(a=1)
        obj['a']=2
        self.assertEqual(2, obj.a)

    def test_getindex(self):
        obj = Obj(a=1)
        self.assertEqual(1,obj['a'])

    def test_reserved_fields(self):
        obj = Obj()
        for field in _EXCEPTIONS:
            self.assertRaises(ValueError,lambda: setattr(obj,field,None))

    def test_keys_values_and_items(self):
        obj = Obj(a=1,b=2,c=3,d=4,e=5,f=6)
        self.assertListEqual(list("abcdef"),list(obj.keys()))
        self.assertListEqual([1,2,3,4,5,6],list(obj.values()))
        self.assertListEqual(
            list(zip("abcdef",[1,2,3,4,5,6])),
            list(obj.items()))

    def test_update(self):
        obj = Obj(a=1,b=2)
        obj = obj.update(b=-2, c=-3)
        self.assertEqual(obj.a,1)
        self.assertEqual(obj.b,-2)
        self.assertEqual(obj.c,-3)

    def test_hasattr(self):
        obj = Obj(a=1)
        self.assertEqual(True,hasattr(obj,'a'))
        self.assertEqual(False,hasattr(obj,'b'))

    def test_remove(self):
        obj = Obj(a=1)
        obj = obj.remove('a')
        self.assertEqual(False,hasattr(obj,'a'))
        self.assertRaises(AttributeError,lambda: print(obj.a))


    def test_repr(self):
        obj = Obj(a=1,b=Obj(c=3,d=4))
        self.assertEqual("{'a': 1, 'b': {'c': 3, 'd': 4}}",obj.__repr__())

    def test_json(self):
        obj = Obj(a=1,b=Obj(c=3,d=4))
        s = json.dumps(obj)
        self.assertEqual('{"a": 1, "b": {"c": 3, "d": 4}}',s)

    def test_obj_create(self):
        obj = Obj.create({'a':[1,2],'b': {'c':1}})
        self.assertListEqual([1,2],obj.a)
        self.assertEqual(1,obj.b.c)






import pandas as pd
from .._common import Obj
from .query_builder import Queryable
from typing import *
from .builder_helpers.generics import TKey, TValue, KeyValuePair

T = TypeVar('T')


def _df_iter(dataframe):
    for row in dataframe.iterrows():
        yield Obj(**row[1].to_dict())


class Query:
    @staticmethod
    def en(en: Iterable[T]) -> Queryable[T]:
        length = None
        if hasattr(en,'__len__'):
            length = len(en)
        return Queryable(en, length)

    @staticmethod
    def dict(dict: Dict[TKey,TValue])->Queryable[KeyValuePair[TKey,TValue]]:
        return Queryable(
            map(lambda z: KeyValuePair(z[0],z[1]), dict.items()),
            len(dict)
        )

    @staticmethod
    def series(series: pd.Series) -> Queryable[KeyValuePair]:
        return Queryable(
            map(lambda z: KeyValuePair(z[0],z[1]), zip(series.index,series)),
            series.shape[0]
        )


    @staticmethod
    def args(*args: T) -> Queryable[T]:
        return Queryable(args, len(args))


    @staticmethod
    def df(dataframe: pd.DataFrame):
        return Queryable(_df_iter(dataframe), dataframe.shape[0])










from yo_core import *
from collections import OrderedDict
import numpy as np

def strjoin(separator: str) -> Callable[[Queryable],str]:
    return lambda en: separator.join(Query.en(en).select(str))


class pairwise(Callable[[Queryable[T]],Queryable[Tuple[T,T]]]):
    def __init__(self):
        pass

    def _make(self, en):
        old = None
        firstTime = True
        for e in en:
            if not firstTime:
                yield (old, e)
            old = e
            if firstTime:
                firstTime = False

    def __call__(self, en):
        return Queryable(self._make(en))


class count_by(Callable[[Queryable[T]],Queryable[KeyValuePair[T,int]]]):
    def __init__(self, selector):
        self.selector = selector

    def __call__(self, en):
        result = OrderedDict()
        for e in en:
            gr = self.selector(e)
            if gr not in result:
                result[gr] = 0
            result[gr] += 1
        return Query.dict(result)


class shuffle(Callable[[Queryable[T]],Queryable[T]]):
    def __init__(self, random_state: Union[np.random.RandomState,int,None,bool]):
        if random_state is None:
            self.random_state = None
        elif isinstance(random_state,bool):
            if random_state:
                self.random_state = np.random.RandomState(np.random.randint(0,1000000))
            else:
                self.random_state = None
        elif isinstance(random_state,int):
            self.random_state = np.random.RandomState(random_state)
        elif isinstance(random_state,np.random.RandomState):
            self.random_state = random_state
        else:
            raise TypeError('random_state must be None, int, bool or RandomState, but was {0}'.format(type(random_state)))

    def _call_iter(self, lst):
        permutation = self.random_state.permutation(len(lst))
        for p in permutation:
            yield lst[p]

    def __call__(self, en):
        if self.random_state is None:
            return en
        lst = np.array(list(en))
        return Queryable(self._call_iter(lst),length=len(lst))



from typing import *
import pandas as pd

class add_ordering_column(Callable[[pd.DataFrame],pd.DataFrame]):
    def __init__(self,
        group_by_columns: Union[str,List[str]],
        order_by_column: Optional[Union[str,Tuple[str,bool]]],
        ordering_column_name = 'order',
        group_id_column_name = None):
        self.group_by_columns = group_by_columns
        self.order_by_column = order_by_column
        self.ordering_column_name = ordering_column_name
        self.group_id_column_name = group_id_column_name

    def __call__(self, df: pd.DataFrame) -> pd.DataFrame:

        if isinstance(self.group_by_columns,str):
            self.group_by_columns = [self.group_by_columns]


        sort_column = list(self.group_by_columns)
        ascendings = [True for z in sort_column]

        if self.order_by_column is not None:
            if isinstance(self.order_by_column,str):
                sort_column.append(self.order_by_column)
                ascendings.append(True)
            elif (
                    isinstance(self.order_by_column,tuple)
                    and len(self.order_by_column)==2
                    and isinstance(self.order_by_column[0],str)
                    and isinstance(self.order_by_column[1],bool)
            ):
                sort_column.append(self.order_by_column[0])
                ascendings.append(self.order_by_column[1])
            else:
                raise ValueError('ordering_by_column expected to be str or Tuple[str,bool], but was {0}'.format(self.order_by_column)) #pragma: no cover

        df = df.sort_values(sort_column,ascending=ascendings)
        sizes = df.groupby(self.group_by_columns).size()
        ordering = [o for sz in sizes for o in range(sz)]
        df = df.assign(**{self.ordering_column_name : ordering})
        if self.group_id_column_name is not None:
            group_id = [index for index, sz in enumerate(sizes) for _ in range(sz)]
            df = df.assign(**{self.group_id_column_name : group_id})
        return df




import os
import json
import pickle
from yo_core._common import Obj
from deprecated import deprecated
from .io import IO

@deprecated('0.4.5', reason='Use yo_extensions.yo.IO class')
def find_root_folder(root_file_name):
    root_path = './'
    for i in range(10):
        if os.path.isfile(root_path+root_file_name):
            return root_path
        root_path+='../'
    raise ValueError("Cound't find the root {1}. The current directory is {0}".format(os.path.abspath('.'), root_file_name))

@deprecated('0.4.5', reason='Use yo_extensions.yo.IO class')
def load_json(filename, as_obj=False):
    result = None
    with open(filename) as file:
        result = json.load(file)
    if as_obj:
        result = Obj.create(result)
    return result

@deprecated('0.4.5', reason='Use yo_extensions.yo.IO class')
def save_json(filename, obj):
    with open(filename,'w') as file:
        json.dump(obj,file,indent=1)

@deprecated('0.4.5', reason='Use yo_extensions.yo.IO class')
def load_pkl(filename):
    with open(filename,'rb') as file:
        return pickle.load(file)


@deprecated('0.4.5', reason='Use yo_extensions.yo.IO class')
def save_pkl(filename, obj):
    with open(filename,'wb') as file:
        pickle.dump(obj,file)


from IPython.display import HTML

def notebook_printable_version(finalize):
    if finalize:
        return HTML('''<script>
        code_show=true; 
        function code_toggle() {
         if (code_show){
         $('div.input').hide();
         $("div[class='prompt output_prompt']").css('visibility','hidden');

         } else {
         $('div.input').show();
         $("div[class='prompt output_prompt']").css('visibility','visible');
         }
         code_show = !code_show
        } 
        $( document ).ready(code_toggle);
        </script>
        <a href="javascript:code_toggle()">Automatically generated report</a>.''')
    else:
        return None

import pandas as pd

def diffset(set1, set2, name1='First',name2='Second'):
    set1=set(set1)
    set2=set(set2)
    return pd.Series([
    len(set1),
    len(set2),
    len(set1-set2),
    len(set2-set1),
    len(set1.intersection(set2)),
    set1==set2,
    set1.issubset(set2),
    set2.issubset(set1),
    ],index=[name1,name2,'{0}-{1}'.format(name1,name2),'{0}-{1}'.format(name2,name1),"Intersection","Match",'LeftIsSubset','RightIsSubset']
)



from typing import *

T = TypeVar('T')
TOut = TypeVar('TOut')
TKey = TypeVar('TKey')
TValue = TypeVar('TValue')

class KeyValuePair(Generic[TKey,TValue]):
    def __init__(self, key: TKey, value: TValue):
        self.key = key
        self.value = value


import unittest
from tests.extensions.common import *
import numpy as np

class MiscMethodsTests(unittest.TestCase):
    def test_pairwise(self):
        result = Query.args(1,2,3).feed(fluq.pairwise()).to_list()
        self.assertListEqual([(1,2),(2,3)],result)

    def test_strjoin(self):
        result = Query.args(1,2,3).feed(fluq.strjoin(','))
        self.assertEqual("1,2,3",result)

    def test_countby(self):
        result = Query.args(1,1,1,2,2,3).feed(fluq.count_by(lambda z: z)).to_series()
        self.assertListEqual([1,2,3],list(result.index))
        self.assertListEqual([3,2,1],list(result))

    def test_shuffle(self):
        arg = Query.en(range(5)).feed(fluq.shuffle(1)).to_list()
        self.assertListEqual([2,1,4,0,3],arg)

    def test_shuffle_rstate(self):
        arg = Query.en(range(5)).feed(fluq.shuffle(np.random.RandomState(1))).to_list()
        self.assertListEqual([2,1,4,0,3],arg)

    def test_shuffle_true(self):
        arg = Query.en(range(5)).feed(fluq.shuffle(True)).to_set()
        self.assertSetEqual({0,1,2,3,4}, arg)
        self.assertEqual(5,len(arg))

    def test_shuffle_false(self):
        res = Query.en(range(5)).feed(fluq.shuffle(False)).to_list()
        self.assertListEqual([0,1,2,3,4],res)


    def test_shuffle_none(self):
        res = Query.en(range(5)).feed(fluq.shuffle(None)).to_list()
        self.assertListEqual([0,1,2,3,4],res)





import unittest
from tests.extensions.common import *

df1 = Query.combinatorics.grid(CatA=['A','B'],CatB=[1,2], CatC=['x','y','z']).to_dataframe()

class OrderingColumnTests(unittest.TestCase):

    def make(self, df, group, order):
        return df.feed(alg.add_ordering_column(group, order, 'order', 'group')).sort_index()

    def check(self, df, group, order, expected_order, group_order):
        result = self.make(df,group,order)
        self.assertListEqual(expected_order, list(result.order))
        self.assertListEqual(group_order, list(result.group))

    def dcheck(self, df, group, order, _, __): #pragma: no cover
        print(self.make(df,group,order))

    def pcheck(self, *args):#pragma: no cover
        pass

    def test_2_grouping_and_sorting_1(self):
        self.check(df1,
            ['CatB','CatC'],
            'CatA',
            [0,0,0,0,0,0,1,1,1,1,1,1],
            [0,1,2,3,4,5,0,1,2,3,4,5]
        )

    def test_2_grouping_and_sorting_2(self):
        self.check(df1,
            ['CatC','CatA'],
            'CatB',
            [0,0,0,1,1,1,0,0,0,1,1,1],
            [0,2,4,0,2,4,1,3,5,1,3,5]
        )

    def test_2_grouping_and_sorting_3(self):
        self.check(df1,
            ['CatA', 'CatB'],
            'CatC',
            [0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2],
            [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3]
        )

    def test_1_grouping_and_sorting_1(self):
        self.check(df1,
            'CatA',
            'CatC',
            [0, 2, 4, 1, 3, 5, 0, 2, 4, 1, 3, 5],
            [0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1]
        )

    def test_1_grouping_and_sorting_2(self):
        self.check(df1,
            'CatB',
            'CatC',
            [0, 2, 4, 0, 2, 4, 1, 3, 5, 1, 3, 5],
            [0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1]
        )


    def test_1_grouping_and_sorting_descending(self):
        self.check(df1,
            'CatB',
            ('CatC',False),
            [4, 2, 0, 4, 2, 0, 5, 3, 1, 5, 3, 1],
            [0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1]
        )


    def test_partial_df(self):
        self.check(df1.iloc[[1,3,4,5,7,8,10,11]],
                    'CatB',
                    'CatA',
                    [0,0,1,2,1,2,3,4],
                    [0,1,1,1,0,0,1,1]
                    )



from tests.extensions.common import *
from pathlib import Path

class TestInitiators(TestCase):
    def test_grid(self):
        df = Query.combinatorics.grid(A=[1, 2], B=['A', 'B']).to_dataframe()
        self.assertListEqual(['A', 'B'], list(df.columns))
        self.assertListEqual([1, 1, 2, 2], list(df.A))
        self.assertListEqual(['A', 'B', 'A', 'B'], list(df.B))


    def test_cartesian(self):
        lst = Query.combinatorics.cartesian([1, 2], ['A', 'B']).to_list()
        self.assertListEqual([
            (1, 'A'),
            (1, 'B'),
            (2, 'A'),
            (2, 'B')
        ], lst)

    def test_triangle(self):
        lst = Query.combinatorics.triangle([1,2,3],True).to_list()
        self.assertListEqual([(1,1),(1,2),(1,3),(2,2),(2,3),(3,3)],lst)

    def test_trianle_no_diagonal(self):
        lst = Query.combinatorics.triangle([1,2,3,4],False).to_list()
        self.assertListEqual([(1,2),(1,3),(1,4),(2,3),(2,4),(3,4)],lst)


    def test_folder(self):
        lst = Query.folder(path('test_fluq_extensions', 'folder')).select(lambda z: z.name).order_by(lambda z: z).to_list()
        self.assertListEqual(['file1', 'file2'], lst)

    def test_folder_as_path(self):
        lst = Query.folder(Path(path('test_fluq_extensions', 'folder'))).select(lambda z: z.name).order_by(
            lambda z: z).to_list()
        self.assertListEqual(['file1', 'file2'], lst)

    def test_folder_wrong_arg(self):
        self.assertRaises(ValueError,lambda:Query.folder(123).to_list())

    def test_folder_nonexisting(self):
        self.assertRaises(ValueError,lambda: Query.folder('NON-EXISTING-FOLDER').to_list())


    def test_powerset(self):
        r = Query.combinatorics.powerset([1,2,3]).to_list()
        self.assertListEqual([
            (),
            (1,),
            (2,),
            (3,),
            (1,2),
            (1,3),
            (2,3),
            (1,2,3)
        ],r)

    def test_powerset_2(self):
        r = Query.combinatorics.powerset(range(20)).distinct().count()
        self.assertEqual(2**20,r)


from typing import *
import math
import copy

class BaseAggregation:
    def reset(self):
        raise NotImplementedError()
    def account(self, element):
        raise NotImplementedError()
    def report(self):
        raise NotImplementedError()
    def get_name(self):
        raise NotImplementedError()
    def process(self, en):
        self.reset()
        for e in en:
            self.account(e)
        return self.report()

class Sum(BaseAggregation):
    def __init__(self):
        self.sum=0
        self.has_elements = False

    def reset(self):
        self.sum=0
        self.has_elements = False

    def account(self, element):
        self.sum+=element
        self.has_elements = True

    def report(self):
        if not self.has_elements:
            return None
        return self.sum

    def get_name(self):
        return 'sum'


class Count(BaseAggregation):
    def __init__(self,as_float : bool =False):
        self.cnt = 0
        self.as_float = as_float
    def reset(self):
        self.cnt = 0

    def account(self, element):
        self.cnt+=1

    def report(self):
        if self.as_float:
            return float(self.cnt)
        else:
            return self.cnt

    def get_name(self):
        return 'count'


class Mean(BaseAggregation):
    def __init__(self):
        self.sum = 0
        self.cnt = 0

    def reset(self):
        self.sum=0
        self.cnt=0

    def account(self, element):
        self.sum+=element
        self.cnt += 1

    def report(self):
        if self.cnt==0:
            return None
        return self.sum/float(self.cnt)

    def get_name(self):
        return 'mean'

class Min(BaseAggregation):
    def __init__(self):
        self.min = None
        self.has_element = False

    def reset(self):
        self.min = None
        self.has_element = False

    def account(self, element):
        self.has_element = True
        if self.min is None or self.min>element:
            self.min = element

    def report(self):
        if not self.has_element:
            return None
        return self.min

    def get_name(self):
        return 'min'


class Max(BaseAggregation):
    def __init__(self):
        self.max = None
        self.has_element = False

    def reset(self):
        self.max = None
        self.has_element = False

    def account(self, element):
        self.has_element = True
        if self.max is None or self.max<element:
            self.max = element

    def report(self):
        if not self.has_element:
            return None
        return self.max

    def get_name(self):
        return 'max'

class Any(BaseAggregation):
    def __init__(self):
        self.any = None

    def reset(self):
        self.any = False

    def account(self, element):
        if element:
            self.any = True

    def report(self):
        return self.any

    def get_name(self):
        return 'any'


class All(BaseAggregation):
    def __init__(self):
        self.all = None

    def reset(self):
        self.all = True

    def account(self, element):
        if not element:
            self.all = False

    def report(self):
        return self.all

    def get_name(self):
        return 'all'


class Std(BaseAggregation):
    def __init__(self):
        self.mean = None
        self.count = None
        self.M2 = None

    def reset(self):
        self.mean = 0
        self.count = 0
        self.M2 = 0

    def account(self, element):
        self.count += 1
        delta = element - self.mean
        self.mean += delta / self.count
        delta2 = element - self.mean
        self.M2 += delta * delta2

    def report(self):
        if self.count<2:
            return None
        return math.sqrt(self.M2/(self.count))

    def get_name(self):
        return 'std'



class ListAggregator(BaseAggregation):
    def __init__(self, *aggs: BaseAggregation):
        self.aggs = aggs
        if len(self.aggs) == 0:
            raise ValueError('No aggregators are provided')

    def reset(self):
        for a in self.aggs:
            a.reset()

    def account(self, element):
        for a in self.aggs:
            a.account(element)

    def report(self):
        if len(self.aggs)==1:
            return self.aggs[0].report()
        return {z.get_name(): z.report() for z in self.aggs}

    def get_name(self):
        return 'ListAggregator'

class _FieldListAggregatorComponent:
    def __init__(self, name: str, selector:Optional[Callable], agg: ListAggregator):
        self.name=name
        if selector is not None:
            self.selector = selector
        else:
            self.selector = lambda z: z[name]
        self.agg = agg


class _FieldListAggregatorPromise:
    def __init__(self, aggregator: 'Aggregator', name: str, selector: Optional[Callable]):
        self.fields_list = [(name,selector)]
        self._aggregator = aggregator

    def to_field(self, name, selector: Optional[Callable] = None):
        self.fields_list.append((name,selector))
        return self

    def apply(self, *aggs: BaseAggregation):
        for pair in self.fields_list:
            self._aggregator._components.append(_FieldListAggregatorComponent(pair[0],pair[1],copy.deepcopy(ListAggregator(*aggs))))
        return self._aggregator



class Aggregator(BaseAggregation):
    def __init__(self):
        self._components = [] # type: List[_FieldListAggregatorComponent]

    def to_field(self, name: str, selector: Optional[Callable]=None):
        return _FieldListAggregatorPromise(self,name,selector)

    def reset(self):
        for c in self._components:
            c.agg.reset()


    def account(self, element):
        for c in self._components:
            c.agg.account(c.selector(element))

    def report(self):
        return {c.name:c.agg.report() for c in self._components}

    def get_name(self):
        return 'Aggregator'






def aggregate_with(en, selector:Optional[Callable] = None, *args: BaseAggregation):
    list_agg = ListAggregator(*args)
    list_agg.reset()
    for e in en:
        if selector is None:
            p = e
        else:
            p = selector(e)
        list_agg.account(p)
    return list_agg.report()










from tests.extensions.common import *
from pathlib import Path

class TestLoop(TestCase):
    def test_loop_inc_endless(self):
        self.assertEqual(
            [1,3,5,7,9],
            Query.loop(1,2).take_while(lambda z: z<10).to_list()
        )

    def test_loop_dec_endless(self):
        self.assertEqual(
            [1,-1,-3],
            Query.loop(1,-2).take_while(lambda z: z>-5).to_list()
        )

    def test_loop_inc_notequal(self):
        self.assertEqual(
            [1,4,7],
            Query.loop(1,3,10).to_list()
        )

    def test_loop_inc_equal(self):
        self.assertEqual(
            [1,4,7,10],
            Query.loop(1,3,10,LoopEndType.Equal).to_list()
        )

    def test_loop_inc_force(self):
        self.assertEqual(
            [1,4,7,9],
            Query.loop(1,3,9,LoopEndType.Force).to_list()
        )


    def test_loop_dec_notequal(self):
        self.assertEqual(
            [1,-2,-5],
            Query.loop(1,-3,-6).to_list()
        )

    def test_loop_dec_equal(self):
        self.assertEqual(
            [1,-2,-5, -8],
            Query.loop(1,-3,-8,LoopEndType.Equal).to_list()
        )

    def test_loop_dec_force(self):
        self.assertEqual(
            [1,-2, -5, -7],
            Query.loop(1,-3,-7,LoopEndType.Force).to_list()
        )

    def test_loop_inc_empty(self):
        self.assertEqual(
            [],
            Query.loop(1,1,1).to_list()
        )

    def test_loop_inc_emptyend(self):
        self.assertEqual(
            [1],
            Query.loop(1, 1, 1, LoopEndType.Equal).to_list()
        )

    def test_loop_inc_emptyendforce(self):
        self.assertEqual(
            [1],
            Query.loop(1, 1, 1, LoopEndType.Force).to_list()
        )

    def test_loop_dec_empty(self):
        self.assertEqual(
            [],
            Query.loop(1,-1,1).to_list()
        )

    def test_loop_dec_emptyend(self):
        self.assertEqual(
            [1],
            Query.loop(1, -1, 1, LoopEndType.Equal).to_list()
        )

    def test_loop_dec_emptyendforce(self):
        self.assertEqual(
            [1],
            Query.loop(1, -1, 1, LoopEndType.Force).to_list()
        )

    def test_loop_inc_contradictory_raise(self):
        self.assertRaises(ValueError,lambda: Query.loop(1,-1,10).to_list())

    def test_loop_dec_contradictory_raise(self):
        self.assertRaises(ValueError,lambda: Query.loop(1,1,-10).to_list())





from tests.core.common import *

class ExtendedMethodsTests(LinqTestBase):
    def test_argmax_1(self):
        self.assertEqual(-1, Query.args(-1, 0, 1).argmax(lambda z: abs(z)))

    def test_argmax_2(self):
        self.assertRaises(ValueError,lambda: Query.args().argmax(lambda z: abs(z)))

    def test_argmax_or_default_1(self):
        self.assertEqual(-1, Query.args(-1, 0, 1).argmax_or_default(lambda z: abs(z)))

    def test_argmax_or_default_2(self):
        self.assertEqual(-10, Query.args().argmax_or_default(lambda z: abs(z), -10))

    def test_argmin_1(self):
        self.assertEqual(0, Query.args(-1, 0, 1).argmin(lambda z: abs(z)))

    def test_argmin_2(self):
        self.assertRaises(ValueError,lambda: Query.args().argmin(lambda z: abs(z)))


    def test_argmin_or_default_1(self):
        self.assertEqual(0, Query.args(-1, 0, 1).argmin_or_default(lambda z: abs(z)))

    def test_argmin_or_default_2(self):
        self.assertEqual(-10, Query.args().argmin_or_default(lambda z: abs(z), -10))

    def test_foreach(self):
        buf = []
        Query.args(1, 2, 3).foreach(buf.append)
        self.assertListEqual([1,2,3],buf)

    def test_foreach_and_continue(self):
        buf = []
        Query.args(1, 2, 3).foreach_and_continue(lambda z: buf.append(z + 10)).foreach(buf.append)
        self.assertListEqual([11,1,12,2,13,3], buf)

    def test_append(self):
        self.assertQuery(Query.args(1, 2).append(5, 6), 1, 2, 5, 6)

    def test_prepend(self):
        self.assertQuery(Query.args(1, 2).prepend(5, 6), 5, 6, 1, 2)


import unittest
from yo_core._common import *
import numpy as np

class TestClass:
    def __init__(self):
        self.number = 123
        self.array = [1,2,3]
        self.matrix = np.array([[1,2],[3,4]])
        self.tensor = np.array([ [[1,2],[3,4]] , [[5,6],[7,8]] ])
        self.dictionary = dict(a=2,b=3)

    def call(self):
        return 'ok'

    def arg_call(self, arg1, arg2):
        return arg1+arg2

    def throwing_call(self):
        raise ValueError('')

class MaybeTests(unittest.TestCase):
    def test_field_access(self):
        self.assertEqual(123,maybe(TestClass(),'number'))

    def test_array_access(self):
        self.assertEqual(2,maybe(TestClass(), 'array', [1]))

    def test_matrix_access(self):
        self.assertEqual(3,maybe(TestClass(), 'matrix', [1,0]))

    def test_3dmatrix_access(self):
        self.assertEqual(8,maybe(TestClass(), 'tensor', [1,1,1]))

    def test_dict_access(self):
        self.assertEqual(3,maybe(TestClass(), 'dictionary', ['b']))

    def test_none(self):
        self.assertIsNone(maybe(None,'a'))

    def test_call_1(self):
        self.assertEqual('ok',maybe(TestClass(),TestClass.call))

    def test_call_2(self):
        self.assertEqual('ok',maybe(TestClass(),'call',mb_call()))

    def test_call_with_args(self):
        self.assertEqual(5,maybe(TestClass(),'arg_call',mb_call(2,3)))

    def test_throwing(self):
        self.assertRaises(ValueError,lambda: maybe(TestClass(),'throwing_call',mb_call()))


    def test_throwing_with_try_catch(self):
        self.assertEqual(None,maybe(TestClass(), 'throwing_call', mb_call().try_catch()))

    def test_default(self):
        self.assertEqual('default',maybe(TestClass(),'non_existing', default='default'))



from setuptools import setup, find_packages

def readme():
      with open('README.md') as file:
            return file.read()

setup(name='yo_ds',
      version='0.4.15',
      description='The toolkit for data science projects with a focus on functional programming',
      long_description=readme(),
      long_description_content_type='text/markdown',
      classifiers = [
            'Development Status :: 3 - Alpha',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3.6',
            'Topic :: Software Development :: Libraries :: Python Modules'
      ],
      url='http://comingsoon.com',
      author='Yuri Okulovsky',
      author_email='yuri.okulovsky@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=[
            'pandas',
            'coverage',
            # the modules below are required for yo_extensions+
            'ipython',
            'ipywidgets',
            'matplotlib',
            'tqdm',
            'seaborn',
            'jsonpickle',
            'deprecated',
            # the modules below are required for yo_ml+
            'sklearn',
            'keras',
            'tensorflow',
            'pydot'
      ],
      test_suite='nose.collector',
      tests_require=['nose'],
      include_package_data = True,
      zip_safe=False)


from tests.core.common import *
import numpy


class AggregationMethodsTests(LinqTestBase):


    def test_count(self):
        self.assertEqual(3,Query.args(1,2,3).count())

    def test_count_float(self):
        t = Query.args(1,2,3).aggregate_with(agg.Count(True))
        self.assertIsInstance(t,float)
        self.assertEqual(3.0,t)

    def test_sum(self):
        self.assertEqual(6,Query.args(1,2,3).sum())

    def test_sum_on_empty(self):
        self.assertIsNone(Query.args().sum())

    def test_min(self):
        self.assertEqual(1,Query.args(1,2,3).min())

    def test_min_on_empty(self):
        self.assertIsNone(Query.args().min())

    def test_max(self):
        self.assertEqual(3,Query.args(1,2,3).max())

    def test_max_on_empty(self):
        self.assertIsNone(Query.args().max())

    def test_mean(self):
        self.assertEqual(2.0,Query.args(1,2,3).mean())

    def test_mean_on_empty(self):
        self.assertIsNone(Query.args().mean())

    def test_std(self):
        for i in range(10):
            r = numpy.random.RandomState(i).rand(10)
            expected = numpy.std(r)
            actual = Query.en(r).aggregate_with(agg.Std())
            self.assertAlmostEqual(expected,actual,5)




    def test_all_1(self):
        self.assertEqual(True, Query.args(1, 2, 3).all(lambda z: z > 0))

    def test_all_2(self):
        self.assertEqual(False, Query.args(1, 2, 3).all(lambda z: z > 1))

    def test_all_3(self):
        self.assertEqual(True, Query.args(dict()).all())

    def test_all_4(self):
        self.assertEqual(True, Query.args().all())


    def test_any_1(self):
        self.assertEqual(True, Query.args(1, 2, 3).any(lambda z: z > 1))

    def test_any_2(self):
        self.assertEqual(False, Query.args(1, 2, 3).any(lambda z: z < 1))

    def test_any_3(self):
        self.assertEqual(True, Query.args(dict()).any())

    def test_any_4(self):
        self.assertEqual(False, Query.args().any())




    def test_aggregate_with(self):
        self.assertDictEqual({'sum':6,'count':3,'mean':2.0},Query.args(1,2,3).aggregate_with(agg.Sum(),agg.Count(),agg.Mean()))


    def test_aggregator(self):
        self.assertDictEqual(
            {
                'a': {
                    'min': 1,
                    'max': 3
                },
                'b': 20.0,
                'sum' : {
                    'min': 11,
                    'mean': 22.0
                }
            },
            Query.args(1,2,3).select(lambda z: Obj(a=z, b=10*z)).aggregate_with(
                agg.Aggregator()
                .to_field('a').apply(agg.Min(),agg.Max())
                .to_field('b').apply(agg.Mean())
                .to_field('sum',lambda z: z['a']+z['b']).apply(agg.Min(),agg.Mean())
            ))

    def test_aggregator_and_to(self):
        self.assertDictEqual({
                'a': {'min':1,'max':3},
                'b': {'min':10, 'max':30}
            },
            Query.args(1,2,3).aggregate_with(agg.Aggregator()
                                             .to_field('a',lambda z: z)
                                             .to_field('b',lambda z: 10*z)
                                             .apply(agg.Min(),agg.Max())))





import pandas as pd
import numpy as np
from yo_core._common import Obj
from .generics import KeyValuePair
from typing import *

def _or_default(with_default, default):
    if with_default:
        return default
    raise ValueError("Sequence contains no elements")


def first(en, with_default, default=None):
    for e in en:
        return e
    return _or_default(with_default, default)

def single(en, with_default, default=None):
    firstTime = True
    value = None
    for e in en:
        if firstTime:
            value = e
            firstTime = False
        else:
            raise ValueError('Sequence contains more than one element')
    if not firstTime:
        return value
    return _or_default(with_default, default)

def last(en, with_default, default=None):
    firstTime = True
    value = None
    for e in en:
        if firstTime:
            firstTime = False
        value = e
    if not firstTime:
        return value
    return _or_default(with_default, default)


def to_set(en):
    result = set()
    for e in en:
        if e in result:
            raise ValueError("Duplicating element {0}".format(e))
        result.add(e)
    return result


def to_ndarray(en):
    return np.array(list(en))

def to_dataframe(q, **kwargs):
    data = list(q)
    df = pd.DataFrame(data,**kwargs)
    if len(data)>0 and isinstance(data[0],Obj) and set(data[0].keys())==set(df.columns): #TODO: 'columns' has to have precedence
        columns = list(data[0].keys())
        df = df[columns]
    return df


class _KeyValueSelector:
    def __init__(self, keySelector: bool, allowNonKeyValue: bool):
        self.isFirstTime = True
        self.isKeyValuePairs = False
        self.keySelector = keySelector
        self.allowNonKeyValue = allowNonKeyValue

    def select(self, e):
        if self.isFirstTime:
            self.isFirstTime=False
            if isinstance(e,KeyValuePair):
                self.isKeyValuePairs = True
            else:
                self.isKeyValuePairs = False
        else:
            if isinstance(e,KeyValuePair)!=self.isKeyValuePairs:
                raise ValueError('Sequence was a mixture od KeyValuePair and something else, which is not allowed in this context')
        if self.isKeyValuePairs:
            if self.keySelector:
                return e.key
            else:
                return e.value
        else:
            if self.allowNonKeyValue:
                return e
            else:
                raise ValueError('Sequence was not of KeyValuePair')


def to_dictionary(en, key_selector, value_selector):
    if (key_selector is None) != (value_selector is None):
        raise ValueError('key_selector and value_selector must be provided either both or none')
    if key_selector is None and value_selector is None:
        key_selector = _KeyValueSelector(True,False).select
        value_selector = _KeyValueSelector(False,False).select
    result = {}
    for e in en:
        key = key_selector(e)
        if key in result:
            raise ValueError('Duplicating value for the key `{0}`'.format(key))
        value = value_selector(e)
        result[key]=value
    return result

def to_series(en, item_to_value, item_to_index):
    index = []
    values = []
    if item_to_value is None:
        item_to_value = _KeyValueSelector(False,True).select

    keySelectorObject = None # type: Optional[_KeyValueSelector]
    if item_to_index is None:
        keySelectorObject = _KeyValueSelector(True,True)
        item_to_index = keySelectorObject.select

    for e in en:
        values.append(item_to_value(e))
        index.append(item_to_index(e))

    if keySelectorObject is not None and not keySelectorObject.isKeyValuePairs:
        index = None

    return pd.Series(values,index)




from yo_core import *
from pathlib import Path
import itertools, gzip, pickle
import os

def _grid_iter(keys,lists):
    for config in itertools.product(*lists):
        yield Obj(**{key: value for key, value in zip(keys, config)})

def _grid(**kwargs):
    keys = list(kwargs)
    lists = [kwargs[key] for key in keys]
    length = 1
    for l in lists:
        length*=len(l)
    return Queryable(_grid_iter(keys, lists), length)


def _grid_args(args):
    length = 1
    for l in args:
        length*=len(l)
    return Queryable(itertools.product(*args),length)



def _triangle_iter(items,with_diagonal):
    for index1,item1 in enumerate(items):
        begin = index1
        if not with_diagonal:
            begin+=1
        for index2 in range(begin,len(items)):
            item2=items[index2]
            yield (item1,item2)




def _text_file(filename, **kwargs):
    with open(filename, 'r', **kwargs) as file:
        for line in file:
            if line.endswith('\n'):
                line = line[:-1]
            yield line

def _zip_text_file(filename, encoding):
    with gzip.open(filename, 'rb') as f:
        for line in f:
            if line.endswith(b'\n'):
                line = line[:-1]
            yield line.decode(encoding)


def _pickle_file(filename):
    with open(filename, 'rb') as file:
        while file.read(1):
            file.seek(-1, 1)
            length = file.read(4)
            length = int.from_bytes(length, 'big')
            dump = file.read(length)
            obj = pickle.loads(dump)
            yield obj



def folder(location: Union[Path, str], pattern: Optional[str] = None):
    if isinstance(location,str):
        location = Path(location)
    elif isinstance(location,Path):
        pass
    else:
        raise ValueError('Location should be either str or Path, but was {0}, {1}'.format(type(location),location))

    if not os.path.isdir(location):
        raise ValueError('{0} is not a directory'.format(location))

    return location.glob(pattern)


class FileQuery:
    def text(self, path: Union[str,Path], **file_kwargs) -> Queryable[str]:
        return Queryable(_text_file(path, **file_kwargs))

    def zipped_text(self, path: Union[str,Path], encoding: str ='utf-8') -> Queryable[str]:
        return Queryable(_zip_text_file(path, encoding))

    def pickle(self, path: Union[str,Path]) -> Queryable[Any]:
        return Queryable(_pickle_file(path))



T = TypeVar('T')

from itertools import combinations

class CombinatoricsQuery:
    def grid(self, **kwargs)->Queryable[Obj]:
        return _grid(**kwargs)

    def cartesian(self,*args)->Queryable[Tuple]:
        return _grid_args(args)

    def triangle(self, items: T, with_diagonal=True) -> Queryable[Tuple[T, T]]:
        return Queryable(_triangle_iter(items, with_diagonal), (len(items) * (len(items) - 1)) // 2)

    def powerset(self, iterable: Iterable[T]) -> Queryable[Tuple]:
        xs = list(iterable)
        return Query.en(range(len(xs) + 1)).select_many(lambda z: combinations(xs, z))

from enum import Enum

class LoopEndType(Enum):
    NotEqual = 0
    Equal = 1
    Force = 2

class loop_maker:
    def __init__(self, begin, delta, end, endType: LoopEndType):
        self.begin = begin
        self.delta = delta
        self.end = end
        self.endType = endType
        self.less_comparison = None

    def _cmp_less_like(self, left, right):
        if self.less_comparison is None:
            raise ValueError('loop_maker error: cmp is requested, but less_comparison was not set')
        if self.less_comparison:
            return left<right
        else:
            return left>right

    def make(self):
        value = self.begin

        if self.end is not None:
            if value==self.end:
                if self.endType != LoopEndType.NotEqual:
                    yield value
                return

        yield value

        value = value + self.delta
        if value > self.begin:
            self.less_comparison = True
        else:
            self.less_comparison = False

        if self.end is not None and not self._cmp_less_like(self.begin,self.end):
            raise ValueError('`end` and `delta` parameters are contradictory. ')

        while True:
            if self.end is None:
                yield value
                value+=self.delta
                continue

            if value == self.end:
                if self.endType != LoopEndType.NotEqual:
                    yield value
                break

            elif self._cmp_less_like(value,self.end):
                yield value
                value+=self.delta
                continue

            else:
                if self.endType == LoopEndType.Force:
                    yield self.end
                break












